unit MListViewEx;

interface

uses
  SysUtils, Classes, Controls, ComCtrls, Types, Windows, Graphics, ExtCtrls, ImgList;

type
  TMViewStyle = (mvsIcon, mvsSmallIcon, mvsList, mvsReport, mvsThumbnail);

  TMLVItemThumbnailEvent = procedure(Sender: TObject; AItem: TListItem;
    AThumbnail: TGraphic) of object;

  TMListViewEx = class(TCustomListView)
  private
    FIsThumbnail : Boolean;
    FTempImages  : TCustomImageList;
    FSelMask     : TBitmap;

    FMaskSelected : Boolean;

    FOnItemThumbnail : TMLVItemThumbnailEvent;

    Procedure InitMask;
    Procedure SwapLargeImages;
    function GetThumbnailHeight: Integer;
    function GetThumbnailWidth: Integer;
    function GetViewStyleEx: TMViewStyle;
    function GetLargeImagesEx: TCustomImageList;
    procedure SetThumbnailHeight(const Value: Integer);
    procedure SetThumbnailWidth(const Value: Integer);
    procedure SetViewStyleEx(const Value: TMViewStyle);
    procedure SetLargeImagesEx(const Value: TCustomImageList);

  protected
    function IsCustomDrawn(Target: TCustomDrawTarget; Stage: TCustomDrawStage): Boolean; Override;
    function CustomDrawItem(Item: TListItem; State: TCustomDrawState;
      Stage: TCustomDrawStage): Boolean; Override;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

  published
    Property MaskSelected    : Boolean Read FMaskSelected      Write FMaskSelected default false;
    Property ThumbnailWidth  : Integer Read GetThumbnailWidth  Write SetThumbnailWidth default 16;
    Property ThumbnailHeight : Integer Read GetThumbnailHeight Write SetThumbnailHeight default 16;
    Property LargeImages     : TCustomImageList Read GetLargeImagesEx Write SetLargeImagesEx;
    Property ViewStyle       : TMViewStyle      Read GetViewStyleEx   Write SetViewStyleEx default mvsIcon;
    Property OnItemThumbnail : TMLVItemThumbnailEvent Read FOnItemThumbnail Write FOnItemThumbnail;

    property Action;
    property Align;
    property AllocBy;
    property Anchors;
    property BevelEdges;
    property BevelInner;
    property BevelOuter;
    property BevelKind default bkNone;
    property BevelWidth;
    property BiDiMode;
    property BorderStyle;
    property BorderWidth;
    property Checkboxes;
    property Color;
    property Columns;
    property ColumnClick;
    property Constraints;
    property Ctl3D;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property FlatScrollBars;
    property FullDrag;
    property GridLines;
    property HideSelection;
    property HotTrack;
    property HotTrackStyles;
    property HoverTime;
    property IconOptions;
    property Items;
    property MultiSelect;
    property OwnerData;
    property OwnerDraw;
    property ReadOnly default False;
    property RowSelect;
    property ParentBiDiMode;
    property ParentColor default False;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowColumnHeaders;
    property ShowWorkAreas;
    property ShowHint;
    property SmallImages;
    property SortType;
    property StateImages;
    property TabOrder;
    property TabStop default True;
    property Visible;
    property OnAdvancedCustomDraw;
    property OnAdvancedCustomDrawItem;
    property OnAdvancedCustomDrawSubItem;
    property OnChange;
    property OnChanging;
    property OnClick;
    property OnColumnClick;
    property OnColumnDragged;
    property OnColumnRightClick;
    property OnCompare;
    property OnContextPopup;
    property OnCustomDraw;
    property OnCustomDrawItem;
    property OnCustomDrawSubItem;
    property OnData;
    property OnDataFind;
    property OnDataHint;
    property OnDataStateChange;
    property OnDblClick;
    property OnDeletion;
    property OnDrawItem;
    property OnEdited;
    property OnEditing;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnGetImageIndex;
    property OnGetSubItemImage;
    property OnDragDrop;
    property OnDragOver;
    property OnInfoTip;
    property OnInsert;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnSelectItem;
    property OnStartDock;
    property OnStartDrag;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('MentalCtrls', [TMListViewEx]);
end;

{ TMListViewEx }

constructor TMListViewEx.Create(AOwner: TComponent);
begin
  inherited;
    FIsThumbnail := false;
    FTempImages  := TImageList.Create( Self );
    FSelMask     := TBitmap.Create;
    InitMask;
    FMaskSelected:= false;
end;

destructor TMListViewEx.Destroy;
begin
    If ( FIsThumbnail ) Then
        SwapLargeImages;
    FTempImages.Free;
    FSelMask.Free;
  inherited;
end;

function TMListViewEx.IsCustomDrawn(Target: TCustomDrawTarget;
  Stage: TCustomDrawStage): Boolean;
begin
    If ( (*FIsThumbnail AND *)( Stage = cdPrePaint ) AND ( Target = dtItem ) ) Then
        Result := true
    Else
        Result := Inherited IsCustomDrawn( Target, Stage );
end;

function TMListViewEx.CustomDrawItem(Item: TListItem;
  State: TCustomDrawState; Stage: TCustomDrawStage): Boolean;

    Procedure StrechGlyph( aIcon : TBitmap );
    Var
        t : TBitmap;
        r : Double;
        rc : TRect;
    Begin
        If ( ( aIcon.Width > Inherited LargeImages.Width )
            OR ( aIcon.Height > Inherited LargeImages.Height ) ) Then
        Begin
            t := TBitmap.Create;
            Try
                t.Assign( aIcon );
                aIcon.Width  := Inherited LargeImages.Width;
                aIcon.Height := Inherited LargeImages.Height;
                aIcon.Canvas.FillRect( Rect( 0, 0, aIcon.Width, aIcon.Height ) );
                r := aIcon.Width / t.Width;
                If ( t.Height * r > aIcon.Height ) Then
                    r := aIcon.Height / t.Height;
                rc.Left   := Round( ( aIcon.Width - t.Width * r ) / 2 );
                rc.Right  := rc.Left + Round( t.Width * r );
                rc.Top    := Round( ( aIcon.Height - t.Height * r ) / 2 );
                rc.Bottom := rc.Top + Round( t.Height * r );
                aIcon.Canvas.StretchDraw( rc, t );
            Finally
                t.Free;
            End;
        End;
    End;

    Function GetIconPoint( aRc : TRect; aIcon : TBitmap ) : TPoint;
    Var
        n : Integer;
    Begin
        Result := aRc.TopLeft;
        n := aRc.Right - aRc.Left - aIcon.Width;
        If ( n > 0 ) Then
            Result.X := aRc.Left + n DIV 2;
        n := aRc.Bottom - aRc.Top - aIcon.Height;
        If ( n > 0 ) Then
            Result.Y := aRc.Top + n DIV 2;
    End;

Var
    glyph : TBitmap;
    rc : TRect;
    ptIcon : TPoint;
    clrText : TColor;
    n : Integer;
begin
    Result := Inherited CustomDrawItem( Item, State, Stage );
    If ( Result ) Then
    Begin  //  Set Canvas's changed flag
        clrText := Canvas.Font.Color;
        Canvas.Font.Color := clWhite;
        Canvas.Font.Color := clrText;
    End;
    If ( Result AND FIsThumbnail ) Then
    Begin
        glyph := TBitmap.Create;
        Try
            FOnItemThumbnail( Self, Item, glyph );
            If ( glyph.Empty ) Then
            Begin
                glyph.Width  := Inherited LargeImages.Width;
                glyph.Height := Inherited LargeImages.Height;
                rc := Rect( 0, 0, glyph.Width, glyph.Height );
                Frame3D( glyph.Canvas, rc, clBlack, clSilver, 2 );
                n := ( glyph.Width + glyph.Height ) DIV 20;
                glyph.Canvas.Pen.Width := n DIV 2;
                glyph.Canvas.Pen.Color := clRed;
                glyph.Canvas.MoveTo( n, n );
                glyph.Canvas.LineTo( glyph.Width - n, glyph.Height - n );
                glyph.Canvas.MoveTo( n, glyph.Height - n );
                glyph.Canvas.LineTo( glyph.Width - n, n );
            End;
            StrechGlyph( glyph );
            rc := Item.DisplayRect( drIcon );
            ptIcon := GetIconPoint( rc, glyph );
            If ( cdsSelected IN State ) Then
            Begin
                Canvas.Brush.Color := clHighlight;
                Canvas.FillRect( rc );
                Canvas.Brush.Color := Color;
                If ( FMaskSelected ) Then
                Begin
                    glyph.Canvas.CopyMode := cmSrcAnd;
                    glyph.Canvas.Draw( 0, 0, FSelMask );
                End;
            End;
            Canvas.Draw( ptIcon.X, ptIcon.Y, glyph );
            If ( cdsSelected IN State ) Then
                Frame3D( Canvas, rc, clBlack, clSilver, 2 )
            Else
                Frame3D( Canvas, rc, clSilver, clBlack, 2 );
        Finally
            glyph.Free;
        End;
    End;
end;

procedure TMListViewEx.InitMask;
Var
    i, j : Integer;
begin
    FSelMask.Width  := FTempImages.Width;
    FSelMask.Height := FTempImages.Height;
    For i := 0 To FSelMask.Height - 1 Do
    Begin
        For j := 0 To FSelMask.Width - 1 Do
        Begin
            If ( ( ( i MOD 2 = 0 ) AND ( j MOD 2 = 1 ) )
                OR ( ( i MOD 2 = 1 ) AND ( j MOD 2 = 0 ) ) ) Then
                FSelMask.Canvas.Pixels[j, i] := clHighlight
            Else
                FSelMask.Canvas.Pixels[j, i] := clWhite;
        End;
    End;
end;

procedure TMListViewEx.SwapLargeImages;
Var
    t : TCustomImageList;
Begin
    t := Inherited LargeImages;
    Inherited LargeImages := FTempImages;
    FTempImages := t;
    FIsThumbnail := Not FIsThumbnail;
End;

function TMListViewEx.GetLargeImagesEx: TCustomImageList;
begin
    If ( FIsThumbnail ) Then
        Result := FTempImages
    Else
        Result := Inherited LargeImages;
end;

function TMListViewEx.GetThumbnailHeight: Integer;
begin
    If ( FIsThumbnail ) Then
        Result := Inherited LargeImages.Width
    Else
        Result := FTempImages.Height;
end;

function TMListViewEx.GetThumbnailWidth: Integer;
begin
    If ( FIsThumbnail ) Then
        Result := Inherited LargeImages.Width
    Else
        Result := FTempImages.Width;
end;

function TMListViewEx.GetViewStyleEx: TMViewStyle;

    Function GetMViewStyle( aVs : TViewStyle ) : TMViewStyle;
    Begin
        Case ( aVs ) Of
        vsSmallIcon :
            Result := mvsSmallIcon;
        vsList :
            Result := mvsList;
        vsReport :
            Result := mvsReport;
        Else
            Result := mvsIcon;
        End;
    End;

begin
    If ( FIsThumbnail ) Then
        Result := mvsThumbnail
    Else
        Result := GetMViewStyle( Inherited ViewStyle );
end;

procedure TMListViewEx.SetLargeImagesEx(const Value: TCustomImageList);
begin
    If ( FIsThumbnail ) Then
    Begin
        If ( Value <> FTempImages ) Then
            FTempImages := Value;
    End
    Else
    Begin
        If ( Value <> Inherited LargeImages ) Then
            Inherited LargeImages := Value;
    End;
end;

procedure TMListViewEx.SetThumbnailHeight(const Value: Integer);
begin
    If ( ( Not FIsThumbnail ) AND ( Value <> FTempImages.Height ) ) Then
    Begin
        FTempImages.Height := Value;
        InitMask;
    End;
end;

procedure TMListViewEx.SetThumbnailWidth(const Value: Integer);
begin
    If ( ( Not FIsThumbnail ) AND ( Value <> FTempImages.Width ) ) Then
    Begin
        FTempImages.Width := Value;
        InitMask;
    End;
end;

procedure TMListViewEx.SetViewStyleEx(const Value: TMViewStyle);

    Function GetViewStyle( aVs : TMViewStyle ) : TViewStyle;
    Begin
        Case ( aVs ) Of
        mvsSmallIcon :
            Result := vsSmallIcon;
        mvsList :
            Result := vsList;
        mvsReport :
            Result := vsReport;
        Else
            Result := vsIcon;
        End;
    End;

begin
    If ( GetViewStyleEx <> Value ) Then
    Begin
        If ( Value = mvsThumbnail ) Then
        Begin
            If ( Assigned( FOnItemThumbnail ) ) Then
                SwapLargeImages;
            If ( Inherited ViewStyle = vsIcon ) Then
              Inherited ViewStyle := vsList;
            Inherited ViewStyle := vsIcon;
        End
        Else
        Begin
            If ( FIsThumbnail ) Then
            Begin
                SwapLargeImages;
                If ( Value = mvsIcon ) Then
                    Inherited ViewStyle := vsList;
            End;
            Inherited ViewStyle := GetViewStyle( Value );
        End;
        Repaint;
        Arrange( arDefault );
    End;
end;

end.
